"""
BetterUptime API Client
"""
from typing import Dict, Union

from betteruptime.api.http_client import HTTPClient
from betteruptime.resources import (
    EscalationPolicy,
    Heartbeat,
    HeartbeatGroup,
    Incident,
    Metadata,
    Monitor,
    MonitorGroup,
    OnCallCalendar,
    StatusPage,
)

BetterUptimeResource = Union[
    EscalationPolicy,
    Heartbeat,
    HeartbeatGroup,
    Incident,
    Metadata,
    Monitor,
    MonitorGroup,
    OnCallCalendar,
    StatusPage,
]


class Client:
    """
    BetterUptime API Client.
    """

    _http_client: HTTPClient
    _resources: Dict[
        str,
        BetterUptimeResource,
    ] = {}

    _heartbeat_groups: HeartbeatGroup
    _heartbeats: Heartbeat
    _incidents: Incident
    _metadata: Metadata
    _monitor_groups: MonitorGroup
    _monitors: Monitor
    _on_calls: OnCallCalendar
    _policies: EscalationPolicy
    _status_pages: StatusPage

    def __init__(self, bearer_token: str) -> None:
        self._http_client = HTTPClient(bearer_token=bearer_token)
        self._heartbeat_groups = HeartbeatGroup(self._http_client)
        self._heartbeats = Heartbeat(self._http_client)
        self._incidents = Incident(self._http_client)
        self._metadata = Metadata(self._http_client)
        self._monitor_groups = MonitorGroup(self._http_client)
        self._monitors = Monitor(self._http_client)
        self._on_calls = OnCallCalendar(self._http_client)
        self._policies = EscalationPolicy(self._http_client)
        self._status_pages = StatusPage(self._http_client)

    @property
    def heartbeat_groups(self) -> HeartbeatGroup:
        r"""BetterUptime heartbeat groupss resource. Returns :class:`HeartbeatGroup` object.

        :rtype: betteruptime.resources.heartbeat_groups.HeartbeatGroup
        """
        return self._heartbeat_groups

    @property
    def heartbeats(self) -> Heartbeat:
        r"""BetterUptime heartbeats resource. Returns :class:`Heartbeat` object.

        :rtype: betteruptime.resources.heartbeats.Heartbeat
        """
        return self._heartbeats

    @property
    def incidents(self) -> Incident:
        r"""BetterUptime incidents resource. Returns :class:`Incident` object.

        :rtype: betteruptime.resources.incidents.Incident
        """
        return self._incidents

    @property
    def metadata(self) -> Metadata:
        r"""BetterUptime metadata resource. Returns :class:`Metadata` object.

        :rtype: betteruptime.resources.metadata.Metadata
        """
        return self._metadata

    @property
    def monitor_groups(self) -> MonitorGroup:
        r"""BetterUptime monitor groups resource. Returns :class:`MonitorGroup` object.

        :rtype: betteruptime.resources.monitor_groups.MonitorGroup
        """
        return self._monitor_groups

    @property
    def monitors(self) -> Monitor:
        r"""BetterUptime monitors resource. Returns :class:`Monitor` object.

        :rtype: betteruptime.resources.monitor.Monitor
        """
        return self._monitors

    @property
    def on_calls(self) -> OnCallCalendar:
        r"""BetterUptime on-call calendar resource. Returns :class:`OnCallCalendar` object.

        :rtype: betteruptime.resources.on_call_calendar.OnCallCalendar
        """
        return self._on_calls

    @property
    def policies(self) -> EscalationPolicy:
        r"""BetterUptime escalation policies resource. Returns :class:`EscalationPolicy` object.

        :rtype: betteruptime.resources.escalation_policies.EscalationPolicy
        """
        return self._policies

    @property
    def status_pages(self) -> StatusPage:
        r"""BetterUptime status pages resource. Returns :class:`StatusPage` object.

        :rtype: betteruptime.resources.status_pages.StatusPage
        """
        return self._status_pages



"""
BetterUptime Incidents Resource
"""
from __future__ import annotations

from betteruptime.api.http_client import HTTPClient
from betteruptime.resources.generic import MutableResource


class Incident(MutableResource):
    """
    Represents BetterUptime Incidents Resource
    """

    def __init__(self, http_client: HTTPClient, name: str = "incidents") -> None:
        super().__init__(http_client, name)

    def __call__(self, resource_id: str) -> Incident:
        new_resource = Incident(http_client=self.http_client)
        new_resource._resource_id = resource_id
        return new_resource



"""
BetterUptime Status Page Reports Resource
"""
from __future__ import annotations

from betteruptime.api.http_client import HTTPClient
from betteruptime.resources.abstract import AbstractResource
from betteruptime.resources.generic import MutableSubResource

from .status_updates import StatusPageStatusUpdates


class StatusPageReport(MutableSubResource):
    """
    Represents BetterUptime Status Page Report Resource.
    """

    _status_updates: StatusPageStatusUpdates

    def __init__(self, http_client: HTTPClient, parent: AbstractResource, name: str = "status-reports") -> None:
        super().__init__(http_client=http_client, parent=parent, name=name)
        self.status_updates = StatusPageStatusUpdates(http_client=http_client, parent=self)

    def __call__(self, resource_id: str) -> StatusPageReport:
        new_resource = StatusPageReport(http_client=self.http_client, parent=self.parent)
        new_resource._resource_id = resource_id
        return new_resource

    @property
    def status_updates(self) -> StatusPageStatusUpdates:
        """
        status_updates property getter.
        """
        return self._status_updates

    @status_updates.setter
    def status_updates(self, status_updates: StatusPageStatusUpdates) -> None:
        """
        status_updates property setter.
        """
        self._status_updates = status_updates



"""
HTTP Client for BetterUptime API client.
"""
# stdlib
import logging
import platform
from threading import Lock
from typing import Any, Dict, Optional

import requests
from yarl import URL

# betteruptime
from betteruptime.api import _API_HOST, _API_MAX_RETRIES, _API_PROXIES, _API_TIMEOUT, _API_VERIFY, _API_VERSION
from betteruptime.api.exceptions import ClientError, HTTPError, HttpTimeout, ProxyError
from betteruptime.util.format import construct_url
from betteruptime.version import version as __version__

logger: logging.Logger = logging.getLogger("betteruptime.api")


def _get_user_agent_header() -> str:
    """
    Compute User-Agent header
    """

    return (
        f"betteruptimepy/{__version__} (python {platform.python_version()};"
        f" os {platform.system().lower()}; arch {platform.machine().lower()})"
    )


def _remove_context(exc: Exception) -> Exception:
    """Python3: remove context from chained exceptions to prevent leaking API keys in tracebacks."""
    exc.__cause__ = None
    return exc


class HTTPClient:
    """
    HTTP client based on 3rd party `requests` module, using a single session.
    This allows us to keep the session alive to spare some execution time.
    """

    _bearer_token: Optional[str] = None
    _headers: Dict[str, str] = {
        "Accept": "application/json",
        "User-Agent": _get_user_agent_header(),
    }
    _session: Optional[requests.Session] = None
    _session_lock: Lock = Lock()

    def __init__(
        self,
        api_url: str = _API_HOST,
        api_version: str = _API_VERSION,
        bearer_token: Optional[str] = None,
    ) -> None:
        self.base_url: URL = URL(api_url.strip("/")) / "api" / api_version.strip("/")
        self._bearer_token = bearer_token
        self._headers.update({"Authorization": f"Bearer {self._bearer_token}"})

    @classmethod
    def request(
        cls,
        method: str,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, Any]] = None,
        json: Any = None,
        timeout: float = _API_TIMEOUT,
        allow_redirects: bool = True,
        proxies: Optional[Dict[str, str]] = _API_PROXIES,
        verify: bool = _API_VERIFY,
        max_retries: int = _API_MAX_RETRIES,
    ) -> requests.Response:
        """
        Sends a request.
        Returns :class:`Response <Response>` object.

        :param method: method for the request.
        :param url: URL for the request.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :type timeout: float or tuple
        :param allow_redirects: (optional) Boolean. Enable/disable GET/OPTIONS/POST/PUT/PATCH/DELETE/HEAD redirection. Defaults to ``True``.
        :type allow_redirects: bool
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param verify: (optional) boolean,  controls whether we verify
            the server's TLS certificate. Defaults to ``True``. When set to
            ``False``, requests will accept any TLS certificate presented by
            the server, and will ignore hostname mismatches and/or expired
            certificates, which will make your application vulnerable to
            man-in-the-middle (MitM) attacks. Setting verify to ``False``
            may be useful during local development or testing.
        :rtype: requests.Response
        """
        try:
            with cls._session_lock:
                if cls._session is None:
                    cls._session = requests.Session()
                    http_adapter = requests.adapters.HTTPAdapter(max_retries=max_retries)
                    cls._session.mount("https://", http_adapter)
                    cls._session.headers.update(cls._headers)

            result = cls._session.request(
                method=method,
                url=url,
                headers=headers,
                params=params,
                json=json,
                timeout=timeout,
                allow_redirects=allow_redirects,
                proxies=proxies,
                verify=verify,
            )

            result.raise_for_status()
        except requests.exceptions.ProxyError as exc:
            raise _remove_context(ProxyError(method, url, exc)) from exc
        except requests.ConnectionError as exc:
            raise _remove_context(ClientError(method, url, exc)) from exc
        except requests.exceptions.Timeout as exc:
            raise _remove_context(HttpTimeout(method, url, timeout)) from exc
        except requests.exceptions.HTTPError as exc:
            if exc.response.status_code in (400, 401, 403, 404, 409, 422, 429):
                # This gets caught afterwards and raises an ApiError exception
                pass
            else:
                raise _remove_context(HTTPError(exc.response.status_code, result.reason)) from exc
        except TypeError as exc:
            raise TypeError(
                "Your installed version of `requests` library seems not compatible with"
                "BetterUptime's usage. We recommend upgrading it ('pip install -U requests')."
            ) from exc

        return result

    def get(self, path: URL, **kwargs: Any) -> requests.Response:
        r"""Sends a GET request. Returns :class:`Response` object.

        :param path: PATH for the request.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        kwargs.setdefault("allow_redirects", True)
        return self.request("GET", construct_url(self.base_url, path), **kwargs)

    def options(self, path: URL, **kwargs: Any) -> requests.Response:
        r"""Sends a OPTIONS request. Returns :class:`Response` object.

        :param path: UPATHRL for the request.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        kwargs.setdefault("allow_redirects", True)
        return self.request("OPTIONS", construct_url(self.base_url, path), **kwargs)

    def head(self, path: URL, **kwargs: Any) -> requests.Response:
        r"""Sends a HEAD request. Returns :class:`Response` object.

        :param path: PATH for the request.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        kwargs.setdefault("allow_redirects", False)
        return self.request("HEAD", construct_url(self.base_url, path), **kwargs)

    def post(self, path: URL, json: Any = None, **kwargs: Any) -> requests.Response:
        r"""Sends a POST request. Returns :class:`Response` object.

        :param path: PATH for the request.
        :param json: (optional) json to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        return self.request("POST", construct_url(self.base_url, path), json=json, **kwargs)

    def put(self, path: URL, json: Any = None, **kwargs: Any) -> requests.Response:
        r"""Sends a PUT request. Returns :class:`Response` object.

        :param path: PATH for the request.
        :param json: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        return self.request("PUT", construct_url(self.base_url, path), json=json, **kwargs)

    def patch(self, path: URL, json: Any = None, **kwargs: Any) -> requests.Response:
        r"""Sends a PATCH request. Returns :class:`Response` object.

        :param path: PATH for the request.
        :param json: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        return self.request("PATCH", construct_url(self.base_url, path), json=json, **kwargs)

    def delete(self, path: URL, **kwargs: Any) -> requests.Response:
        r"""Sends a DELETE request. Returns :class:`Response` object.

        :param path: PATH for the request.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        return self.request("DELETE", construct_url(self.base_url, path), **kwargs)



"""
BetterUptime Status Page Resources Resource
"""
from __future__ import annotations

from betteruptime.api.http_client import HTTPClient
from betteruptime.resources.abstract import AbstractResource
from betteruptime.resources.generic import MutableSubResource


class StatusPageResource(MutableSubResource):
    """
    Represents BetterUptime Status Page Resources Resource.
    """

    def __init__(self, http_client: HTTPClient, parent: AbstractResource, name: str = "resources") -> None:
        super().__init__(http_client=http_client, parent=parent, name=name)

    def __call__(self, resource_id: str) -> StatusPageResource:
        new_resource = StatusPageResource(http_client=self.http_client, parent=self.parent)
        new_resource._resource_id = resource_id
        return new_resource



"""
BetterUptime Resources
"""
from betteruptime.resources.escalation_policies import EscalationPolicy
from betteruptime.resources.heartbeat_groups import HeartbeatGroup
from betteruptime.resources.heartbeats import Heartbeat
from betteruptime.resources.incidents import Incident
from betteruptime.resources.metadata import Metadata
from betteruptime.resources.monitor_groups import MonitorGroup
from betteruptime.resources.monitors import Monitor
from betteruptime.resources.on_call_calendar import OnCallCalendar
from betteruptime.resources.status_pages import StatusPage

__all__ = [
    "EscalationPolicy",
    "HeartbeatGroup",
    "Heartbeat",
    "Incident",
    "Metadata",
    "MonitorGroup",
    "Monitor",
    "OnCallCalendar",
    "StatusPage",
]



"""
Client tests
"""
import types

import pytest
from pytest_mock import MockerFixture

import betteruptime
from betteruptime.api.exceptions import ApiError

pytestmark = pytest.mark.vcr


class TestClient:
    """
    BetterUptime API Client tests
    """

    def test_bearer_token_set(self, client: betteruptime.Client) -> None:
        """
        Test if Bearer token is set
        """
        assert "Authorization" in client.status_pages.http_client._headers
        assert client.status_pages.http_client._headers["Authorization"] == "Bearer fake"

    def test_401(self, client: betteruptime.Client) -> None:
        """
        Test that an invalid token raises an ApiError
        """
        with pytest.raises(ApiError) as excinfo:
            client.status_pages.list()
        assert 401 == excinfo.value.status_code

    def test_list_heartbeat_groups(self, client: betteruptime.Client) -> None:
        """
        Test list heartbeat groups.
        """
        heartbeat_groups = client.heartbeat_groups.list()
        assert isinstance(heartbeat_groups, dict)
        assert "data" in heartbeat_groups
        assert isinstance(heartbeat_groups["data"], list)
        assert len(heartbeat_groups["data"]) == 0

    def test_list_heartbeats(self, client: betteruptime.Client) -> None:
        """
        Test list heartbeats.
        """
        heartbeats = client.heartbeats.list()
        assert isinstance(heartbeats, dict)
        assert "data" in heartbeats
        assert isinstance(heartbeats["data"], list)
        assert len(heartbeats["data"]) == 0

    def test_list_incidents(self, client: betteruptime.Client) -> None:
        """
        Test list incidents.
        """
        incidents = client.incidents.list()
        assert isinstance(incidents, dict)
        assert "data" in incidents
        assert isinstance(incidents["data"], list)
        assert len(incidents["data"]) == 0

    def test_list_metadata(self, client: betteruptime.Client) -> None:
        """
        Test list metadata.
        """
        metadata = client.metadata.list()
        assert isinstance(metadata, dict)
        assert "data" in metadata
        assert isinstance(metadata["data"], list)
        assert len(metadata["data"]) == 0

    def test_list_monitors(self, client: betteruptime.Client) -> None:
        """
        Test list monitors.
        """
        monitors = client.monitors.list()
        assert isinstance(monitors, dict)
        assert "data" in monitors
        assert isinstance(monitors["data"], list)
        assert len(monitors["data"]) == 1
        assert "id" in monitors["data"][0]
        assert monitors["data"][0]["id"] == "123456"
        assert "type" in monitors["data"][0]
        assert monitors["data"][0]["type"] == "monitor"

    def test_list_monitor_groups(self, client: betteruptime.Client) -> None:
        """
        Test list monitor groups.
        """
        monitor_groups = client.monitor_groups.list()
        assert isinstance(monitor_groups, dict)
        assert "data" in monitor_groups
        assert isinstance(monitor_groups["data"], list)
        assert len(monitor_groups["data"]) == 1
        assert "id" in monitor_groups["data"][0]
        assert monitor_groups["data"][0]["id"] == "123456"
        assert "type" in monitor_groups["data"][0]
        assert monitor_groups["data"][0]["type"] == "monitor_group"

    def test_list_monitor_group_monitors(self, client: betteruptime.Client) -> None:
        """
        Test list monitor group monitors.
        """
        monitors = client.monitor_groups("123456").monitors()
        assert isinstance(monitors, dict)
        assert "data" in monitors
        assert isinstance(monitors["data"], list)
        assert len(monitors["data"]) == 1
        assert "id" in monitors["data"][0]
        assert monitors["data"][0]["id"] == "123456"
        assert "type" in monitors["data"][0]
        assert monitors["data"][0]["type"] == "monitor"
        monitors = client.monitor_groups("123456").monitors_iter()
        assert isinstance(monitors, types.GeneratorType)

    def test_list_on_calls(self, client: betteruptime.Client) -> None:
        """
        Test list on_calls.
        """
        on_calls = client.on_calls.list()
        assert isinstance(on_calls, dict)
        assert "data" in on_calls
        assert isinstance(on_calls["data"], list)
        assert len(on_calls["data"]) == 1
        assert "id" in on_calls["data"][0]
        assert on_calls["data"][0]["id"] == "123456"
        assert "type" in on_calls["data"][0]
        assert on_calls["data"][0]["type"] == "on_call_calendar"

    def test_list_policies(self, client: betteruptime.Client) -> None:
        """
        Test list escalation policies.
        """
        policies = client.policies.list()
        assert isinstance(policies, dict)
        assert "data" in policies
        assert isinstance(policies["data"], list)
        assert len(policies["data"]) == 0

    def test_list_status_pages(self, client: betteruptime.Client) -> None:
        """
        Test list status pages.
        """
        status_pages = client.status_pages.list()
        assert isinstance(status_pages, dict)
        assert "data" in status_pages
        assert isinstance(status_pages["data"], list)
        assert len(status_pages["data"]) == 1
        assert "id" in status_pages["data"][0]
        assert status_pages["data"][0]["id"] == "123456"
        assert "type" in status_pages["data"][0]
        assert status_pages["data"][0]["type"] == "status_page"

    def test_list_status_page_reports(self, client: betteruptime.Client) -> None:
        """
        Test list status page reports.
        """
        reports = client.status_pages("123456").reports.list()
        assert isinstance(reports, dict)
        assert "data" in reports
        assert isinstance(reports["data"], list)
        assert len(reports["data"]) == 0
        assert "pagination" in reports
        assert isinstance(reports["pagination"], dict)
        assert "first" in reports["pagination"]
        assert (
            "https://betteruptime.com/api/v2/status-pages/123456/status-reports?page=1"
            == reports["pagination"]["first"]
        )

    def test_list_status_page_report_status_updates(self, client: betteruptime.Client) -> None:
        """
        Test list status page report status updates.
        """
        status_updates = client.status_pages("123456").reports("123456").status_updates.list()
        assert isinstance(status_updates, dict)
        assert "data" in status_updates
        assert isinstance(status_updates["data"], list)
        assert len(status_updates["data"]) == 2
        assert "id" in status_updates["data"][0]
        assert status_updates["data"][0]["id"] == "123456"
        assert "type" in status_updates["data"][0]
        assert status_updates["data"][0]["type"] == "status_update"
        assert "attributes" in status_updates["data"][0]
        assert isinstance(status_updates["data"][0]["attributes"], dict)
        assert "message" in status_updates["data"][0]["attributes"]
        assert status_updates["data"][0]["attributes"]["message"] == "First status update message"
        assert "status_report_id" in status_updates["data"][0]["attributes"]
        assert status_updates["data"][0]["attributes"]["status_report_id"] == 123456
        assert "id" in status_updates["data"][1]
        assert status_updates["data"][1]["id"] == "123457"
        assert "type" in status_updates["data"][1]
        assert status_updates["data"][1]["type"] == "status_update"
        assert "attributes" in status_updates["data"][1]
        assert isinstance(status_updates["data"][1]["attributes"], dict)
        assert "message" in status_updates["data"][1]["attributes"]
        assert status_updates["data"][1]["attributes"]["message"] == "Second status update message"
        assert "status_report_id" in status_updates["data"][1]["attributes"]
        assert status_updates["data"][1]["attributes"]["status_report_id"] == 123456

    def test_list_status_page_resources(self, client: betteruptime.Client) -> None:
        """
        Test list status page resources.
        """
        resources = client.status_pages("123456").resources.list()
        assert isinstance(resources, dict)
        assert "data" in resources
        assert isinstance(resources["data"], list)
        assert len(resources["data"]) == 2
        assert "id" in resources["data"][0]
        assert resources["data"][0]["id"] == "123456"
        assert "type" in resources["data"][0]
        assert resources["data"][0]["type"] == "status_page_resource"
        assert "attributes" in resources["data"][0]
        assert isinstance(resources["data"][0]["attributes"], dict)
        assert "status_page_section_id" in resources["data"][0]["attributes"]
        assert resources["data"][0]["attributes"]["status_page_section_id"] == 1
        assert "resource_id" in resources["data"][0]["attributes"]
        assert resources["data"][0]["attributes"]["resource_id"] == 1
        assert "resource_type" in resources["data"][0]["attributes"]
        assert resources["data"][0]["attributes"]["resource_type"] == "EmailIntegration"
        assert "id" in resources["data"][1]
        assert resources["data"][1]["id"] == "123457"
        assert "type" in resources["data"][1]
        assert resources["data"][1]["type"] == "status_page_resource"
        assert "attributes" in resources["data"][1]
        assert isinstance(resources["data"][1]["attributes"], dict)
        assert "status_page_section_id" in resources["data"][1]["attributes"]
        assert resources["data"][1]["attributes"]["status_page_section_id"] == 1
        assert "resource_id" in resources["data"][1]["attributes"]
        assert resources["data"][1]["attributes"]["resource_id"] == 2
        assert "resource_type" in resources["data"][1]["attributes"]
        assert resources["data"][1]["attributes"]["resource_type"] == "Monitor"

    def test_list_status_page_sections(self, client: betteruptime.Client) -> None:
        """
        Test list status page sections.
        """
        sections = client.status_pages("123456").sections.list()
        assert isinstance(sections, dict)
        assert "data" in sections
        assert isinstance(sections["data"], list)
        assert len(sections["data"]) == 1
        assert "id" in sections["data"][0]
        assert sections["data"][0]["id"] == "123456"
        assert "type" in sections["data"][0]
        assert sections["data"][0]["type"] == "status_page_section"

    def test_get_heartbeat_200(self, client: betteruptime.Client) -> None:
        """
        Test get single heartbeat.
        """
        heartbeat = client.heartbeats.get("123456")
        assert isinstance(heartbeat, dict)
        assert "id" in heartbeat["data"]
        assert heartbeat["data"]["id"] == "123456"
        assert "type" in heartbeat["data"]
        assert heartbeat["data"]["type"] == "heartbeat"

    def test_get_heartbeat_404(self, client: betteruptime.Client) -> None:
        """
        Test get single heartbeat but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.heartbeats.get("123456")
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'heartbeats' resource." == excinfo.value.message
        )

    def test_get_heartbeat_group_200(self, client: betteruptime.Client) -> None:
        """
        Test get single heartbeat group.
        """
        heartbeat_group = client.heartbeat_groups.get("123456")
        assert isinstance(heartbeat_group, dict)
        assert "id" in heartbeat_group["data"]
        assert heartbeat_group["data"]["id"] == "123456"
        assert "type" in heartbeat_group["data"]
        assert heartbeat_group["data"]["type"] == "heartbeat_group"

    def test_get_heartbeat_group_404(self, client: betteruptime.Client) -> None:
        """
        Test get single heartbeat group but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.heartbeat_groups.get("123456")
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'heartbeat-groups' resource." == excinfo.value.message
        )

    def test_get_incident_200(self, client: betteruptime.Client) -> None:
        """
        Test get single incident.
        """
        incident = client.incidents.get("123456")
        assert isinstance(incident, dict)
        assert "id" in incident["data"]
        assert incident["data"]["id"] == "123456"
        assert "type" in incident["data"]
        assert incident["data"]["type"] == "incident"

    def test_get_incident_404(self, client: betteruptime.Client) -> None:
        """
        Test get single incident but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.incidents.get("123456")
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'incidents' resource." == excinfo.value.message
        )

    def test_get_metadata_200(self, client: betteruptime.Client) -> None:
        """
        Test get single metadata.
        """
        metadata = client.metadata.get("123456")
        assert isinstance(metadata, dict)
        assert "id" in metadata["data"]
        assert metadata["data"]["id"] == "123456"
        assert "type" in metadata["data"]
        assert metadata["data"]["type"] == "metadata"

    def test_get_metadata_404(self, client: betteruptime.Client, mocker: MockerFixture) -> None:
        """
        Test get single metadata but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.metadata.get("123456")
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'metadata' resource." == excinfo.value.message
        )

    def test_get_monitor_200(self, client: betteruptime.Client) -> None:
        """
        Test get single monitor.
        """
        monitor = client.monitors.get("123456")
        assert isinstance(monitor, dict)
        assert "id" in monitor["data"]
        assert monitor["data"]["id"] == "123456"
        assert "type" in monitor["data"]
        assert monitor["data"]["type"] == "monitor"

    def test_get_monitor_404(self, client: betteruptime.Client) -> None:
        """
        Test get single monitor but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.monitors.get("123456")
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'monitors' resource." == excinfo.value.message
        )

    def test_get_monitor_group_200(self, client: betteruptime.Client) -> None:
        """
        Test get single monitor group.
        """
        monitor_group = client.monitor_groups.get("123456")
        assert isinstance(monitor_group, dict)
        assert "id" in monitor_group["data"]
        assert monitor_group["data"]["id"] == "123456"
        assert "type" in monitor_group["data"]
        assert monitor_group["data"]["type"] == "monitor_group"

    def test_get_monitor_group_404(self, client: betteruptime.Client) -> None:
        """
        Test get single monitor group but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.monitor_groups.get("123456")
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'monitor-groups' resource." == excinfo.value.message
        )

    def test_get_on_call_200(self, client: betteruptime.Client) -> None:
        """
        Test get single on call calendar.
        """
        on_call = client.on_calls.get("123456")
        assert isinstance(on_call, dict)
        assert "id" in on_call["data"]
        assert on_call["data"]["id"] == "123456"
        assert "type" in on_call["data"]
        assert on_call["data"]["type"] == "on_call_calendar"

    def test_get_on_call_404(self, client: betteruptime.Client) -> None:
        """
        Test get single on call calendar but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.on_calls.get("123456")
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'on-calls' resource." == excinfo.value.message
        )

    def test_get_policy_200(self, client: betteruptime.Client) -> None:
        """
        Test get single escalation policy.
        """
        policy = client.policies.get("123456")
        assert isinstance(policy, dict)
        assert "id" in policy["data"]
        assert policy["data"]["id"] == "123456"
        assert "type" in policy["data"]
        assert policy["data"]["type"] == "policy"

    def test_get_policy_404(self, client: betteruptime.Client) -> None:
        """
        Test get single escalation policy but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.policies.get("123456")
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'policies' resource." == excinfo.value.message
        )

    def test_get_status_page_200(self, client: betteruptime.Client) -> None:
        """
        Test get single status page.
        """
        status_page = client.status_pages.get("123456")
        assert isinstance(status_page, dict)
        assert "id" in status_page["data"]
        assert status_page["data"]["id"] == "123456"
        assert "type" in status_page["data"]
        assert status_page["data"]["type"] == "status_page"

    def test_get_status_page_404(self, client: betteruptime.Client) -> None:
        """
        Test get single status page but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.status_pages.get("123456")
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'status-pages' resource." == excinfo.value.message
        )

    def test_get_status_page_report_200(self, client: betteruptime.Client) -> None:
        """
        Test get single status page report.
        """
        report = client.status_pages("123456").reports("123456").get()
        assert isinstance(report, dict)
        assert "id" in report["data"]
        assert report["data"]["id"] == "123456"
        assert "type" in report["data"]
        assert report["data"]["type"] == "status_report"

    def test_get_status_page_report_404(self, client: betteruptime.Client) -> None:
        """
        Test get single status page report but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.status_pages("123456").reports("123456").get()
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'status-reports' resource." == excinfo.value.message
        )

    def test_get_status_page_report_status_update_200(self, client: betteruptime.Client) -> None:
        """
        Test get single status page report status update.
        """
        status_update = client.status_pages("123456").reports("123456").status_updates.get("123456")
        assert isinstance(status_update, dict)
        assert "id" in status_update["data"]
        assert status_update["data"]["id"] == "123456"
        assert "type" in status_update["data"]
        assert status_update["data"]["type"] == "status_update"

    def test_get_status_page_report_status_update_404(self, client: betteruptime.Client) -> None:
        """
        Test get single status page report status update but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.status_pages("123456").reports("123456").status_updates.get("123456")
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'status-updates' resource." == excinfo.value.message
        )

    def test_get_status_page_resource_200(self, client: betteruptime.Client) -> None:
        """
        Test get single status page resource.
        """
        resource = client.status_pages("123456").resources("123456").get()
        assert isinstance(resource, dict)
        assert "id" in resource["data"]
        assert resource["data"]["id"] == "123456"
        assert "type" in resource["data"]
        assert resource["data"]["type"] == "status_page_resource"

    def test_get_status_page_resource_404(self, client: betteruptime.Client) -> None:
        """
        Test get single status page resource but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.status_pages("123456").resources("123456").get()
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'resources' resource." == excinfo.value.message
        )

    def test_get_status_page_section_200(self, client: betteruptime.Client) -> None:
        """
        Test get single status page section.
        """
        section = client.status_pages("123456").sections("123456").get()
        assert isinstance(section, dict)
        assert "id" in section["data"]
        assert section["data"]["id"] == "123456"
        assert "type" in section["data"]
        assert section["data"]["type"] == "status_page_section"

    def test_get_status_page_section_404(self, client: betteruptime.Client) -> None:
        """
        Test get single status page section but not found.
        """
        with pytest.raises(ApiError) as excinfo:
            client.status_pages("123456").sections("123456").get()
        assert 404 == excinfo.value.status_code
        assert (
            "BetterUptime returned the following HTTP response code: 404"
            " - Not Found while querying 'sections' resource." == excinfo.value.message
        )

    def test_post_monitor_201(self, client: betteruptime.Client) -> None:
        """
        Test create a monitor.
        """
        new_monitor = {
            "monitor_type": "status",
            "url": "https://facebook.com",
            "pronounceable_name": "Facebook homepage",
            "email": True,
            "sms": False,
            "call": False,
            "check_frequency": 30,
            "request_headers": [{"name": "X-Custom-Header", "value": "custom header value"}],
        }
        monitor = client.monitors.create(new_monitor)
        assert isinstance(monitor, dict)
        assert "id" in monitor["data"]
        assert monitor["data"]["id"] == "123456"
        assert "type" in monitor["data"]
        assert monitor["data"]["type"] == "monitor"

    def test_patch_monitor_200(self, client: betteruptime.Client) -> None:
        """
        Test update a monitor.
        """
        changed_fields = {"email": False}
        monitor = client.monitors("123456").update(changed_fields)
        assert isinstance(monitor, dict)
        assert "id" in monitor["data"]
        assert monitor["data"]["id"] == "123456"
        assert "type" in monitor["data"]
        assert monitor["data"]["type"] == "monitor"
        assert "attributes" in monitor["data"]
        assert "email" in monitor["data"]["attributes"]
        assert monitor["data"]["attributes"]["email"] is False

    def test_delete_monitor_204(self, client: betteruptime.Client) -> None:
        """
        Test delete a monitor.
        """
        client.monitors.delete("123456")



"""
API & HTTP Clients exceptions.
"""


class BetterUptimeException(Exception):
    """
    Base class for BetterUptime API exceptions.  Use this for patterns like the following:
        try:
            # do something with the BetterUptime API
        except betteruptime.api.exceptions.BetterUptimeException:
            # handle any BetterUptime-specific exceptions
    """


class ProxyError(BetterUptimeException):
    """
    HTTP connection to the configured proxy server failed.
    """

    def __init__(self, method: str, url: str, exception: Exception):
        message = (
            f"Could not request {method} {url}: Unable to connect to proxy. "
            "Please check the proxy configuration and try again."
        )
        super().__init__(message)


class ClientError(BetterUptimeException):
    """
    HTTP connection to BetterUptime endpoint is not possible.
    """

    def __init__(self, method: str, url: str, exception: Exception):
        message = (
            f"Could not request {method} {url}: {exception}. "
            "Please check the network connection or try again later. "
        )
        super().__init__(message)


class HttpTimeout(BetterUptimeException):
    """
    HTTP connection timeout.
    """

    def __init__(self, method: str, url: str, timeout: float):
        message = (
            f"{method} {url} timed out after {timeout}. "
            "Please try again later. "
            "If the problem persists, please contact support@BetterUptimehq.com"
        )
        super().__init__(message)


class HttpBackoff(BetterUptimeException):
    """
    Backing off after too many timeouts.
    """

    def __init__(self, backoff_period: float):
        message = f"Too many timeouts. Won't try again for {backoff_period} seconds. "
        super().__init__(message)


class HTTPError(BetterUptimeException):
    """
    BetterUptime returned a HTTP error.
    """

    def __init__(self, status_code: int = 400, reason: str = ""):
        reason = f" - {reason}" if reason else ""
        message = f"BetterUptime returned a bad HTTP response code: {status_code}{reason}. " "Please try again later. "

        super().__init__(message)


class ApiError(BetterUptimeException):
    """
    BetterUptime returned an API error (known HTTPError).
    Matches the following status codes: 400, 401, 403, 404, 409, 422, 429.
    """

    def __init__(
        self,
        resource: str,
        status_code: int = 400,
        reason: str = "",
        message: str = "",
    ):
        reason = f" - {reason}" if reason else ""
        message = message if message else ""
        message = (
            f"BetterUptime returned the following HTTP response code: "
            f"{status_code}{reason} "
            f"while querying '{resource}' resource."
            f"{message}"
        )

        super().__init__(message)
        self.resource = resource
        self.status_code = status_code
        self.reason = reason
        self.message = message
